#include <Arduino.h>
#include "main.h"
/*
 * CAN HOME BASE copy from CAN Utility NODE
 * MK's Smart Home Node, universal standard node  
 * CAN bus part based on Henry's Bench CAN Network Demo
 * other code partly based on library demos
 * Started by MK Dez. 2019
 * (c) 2020 by Rainer Müller-Knoche * TO DO: 
 * ! settime request do we have to confirm
 * broadcast pressure not yet implemented
 * power failure for timed TO BE IMPLEMENTED
 * ifpower failure while running timed port will be set but no timer
 * think best for the moment will be we remember if timed and switch off
 * we should send feedback if on after power failure
 * last modified 01. Oktober 2020
*/
//#define DEBUG
//#define I2C
//#define MONITOR
//#define NO_BC
//#define SHOW_CONF
//#define SHOW_STATE
//#define LCDMON
#define VERSION F("0.3.00")
#define COPYRIGHT F("(c) Rainer 2020")
#define DHT_FAIL F("Failed to read from DHT sensor!")
//#define EEPROM_ADDRESS  0x57    // default on RTC

#include <DS3232RTC.h> // https://github.com/JChristensen/DS3232RTC
                       // use address 0x68 ist defines in DS3232.ccp
#include <DHT.h>       // In case we have a sensor
#include <EEPROM.h>    // We will store configuration in EEPROM
//#include <Eeprom24C32_64.h>     // external EEPROM on RTC
#include <SPI.h>               // Needed for comunication
#include <mcp_can.h>           // CAN Bus uint16_terface modified for 8 MHz cristal
#include <OneButton.h>         // Different button functions
#include <timer.h>             // Timmer
#include <timerManager.h>      // for timer
#include <LiquidCrystal_I2C.h> // For optional LCD Display with I2C Interface
#include <PCF8574.h>           // Required for PCF8574
#include <avr/wdt.h>           // watchdog timer
#include "CAN_MK.h"            // all constants

#define DHTPIN 3 // move to CAN_MK.h ?
#define DHTTYPE DHT22

//static Eeprom24C32_64 eeprom(EEPROM_ADDRESS); // future use
const uint8_t RTC_1HZ_PIN(3);   // RTC provides a 1Hz uint16_terrupt signal on this pin
// RAM
// CAN bus                      // See CAN documentation for packet build
unsigned long rxID;                                       // encoded sender ID
unsigned long txID;                                       // encoded transmiitter ID
uint8_t len = 0;                                          // Length of CAN message, 8 uint8_ts max
uint8_t rxHead[4];                                        // decoded rxID source_ID, targe_type, message_type, target_ID
uint8_t txHead[4];                                        // unencode txID source_ID, targe_type, message_type, target_ID
uint8_t rxBuf[8];                                         // CAN receive message buffer, data only
uint8_t txBuf[8];                                         // CAN send message buffer, data only
uint8_t myID = 0;                                         // copy from EEPROM
uint8_t myGroup = 0;                                      // copy from EEPROM
uint16_t sizeOfConf = 0;                                  // to be calculated
uint8_t lcdLine[20];                                      // Line buffer for LCD display (one line)
uint8_t lcdRows = EEPROM.read(displayBase + displayRows); // set from EEPROM
uint8_t lcdCols = EEPROM.read(displayBase + displayCols); // set from EEPROM
uint8_t lcdAddr = EEPROM.read(displayBase + displayAddr); // set from EEPROM
// Mass data transfer handshake
uint16_t dataRxCount = 0;            // amount of receive data,
uint16_t dataRxStore = 0;            // store it
boolean dataRx = false;              // data receive in progress
uint8_t dataRxBuffer[8];             // receive data buffer
uint16_t dataRxIndex = 0;            // index in receive buffer
const uint16_t dataRxTimeout = 5000; // receive timeout
uint8_t dataRxType = 0;              // remenber type of recived Data
uint8_t dataRxSubType = 0;           // Remenber SubType of recived Data
// Sending Data
uint16_t dataTxCount = 0;            // amount of data to send
uint16_t dataTxIndex = 0;            // index to TX data
uint8_t txDestination = 0;           // remeber to whom we sent
boolean dataTx = false;              // data send in progress
boolean dataTxAck = false;           // send data ack
const uint16_t dataTxTimeout = 5000; // send timeout
uint8_t longTimers[32];              // long Time Counter seconds
                                     // index is port time, source, key
uint8_t longState[32];               // ative yes/no
// PIR Sensor
uint8_t pirState = LOW; // PIR state
uint8_t pirValue = 0;   //
uint8_t inCount = 0;
// DHT Sensor
DHT dht(DHTPIN, DHTTYPE);
// RTC
// DS3231_Simple Clock;          // done by lib
// DateTime MyDateAndTime;
// Port Expander I2C address is set in init from EEPROM
PCF8574 expander1;
PCF8574 expander2;
PCF8574 expander3;
PCF8574 expander4;
// LCD
LiquidCrystal_I2C lcd(lcdAddr, lcdCols, lcdRows); // setup the LCD settings set from EEPROM
// Init push buttons name(pin, activ)
uint16_t activeLow = 1; // Button close to GND when pushed
boolean pullUp = true;  // Switch on pullUp is default set
/*
OneButton P1(EEPROM.read(inputBase + 0), activeLow);  // P1(port from EEPROM, active mode)
OneButton P2(EEPROM.read(inputBase + 23), activeLow);  // P2(port from EEPROM, active mode)
OneButton P3(EEPROM.read(inputBase + 46), activeLow);  // P3(port from EEPROM, active mode)
OneButton P4(EEPROM.read(inputBase + 69), activeLow);  // P4(port from EEPROM, active mode)
*/
OneButton P1(button_1, activeLow); // P1(port from EEPROM, active mode)
OneButton P2(button_2, activeLow); // P2(port from EEPROM, active mode)
OneButton P3(button_3, activeLow); // P3(port from EEPROM, active mode)
OneButton P4(button_4, activeLow); // P4(port from EEPROM, active mode)
// Timers
Timer timer;   // for send receive
Timer DHTtemp; // broadcast temperature
Timer DHThum;  // broadcast hmidity
Timer RTCtime; // broadcast time
// CAN from CAN_BUS_Shield-master, modified
MCP_CAN CAN(SPI_CS_PIN); // Set CS pin fixed defined in CAN_MK.h

/* 
 * generate CAN ID 
 * we build the CAN ID with 8 bit sourceID, 2 bit target_type,
 * 4 bit message type, 8 bit targetID, total 22 bit
 * we have 7 bit left to 29 bit of extendet frame for future use
 * NOTE PARAMETER NOT IN SAME ORDER
 */
uint32_t encodeID(uint8_t source,       // source ID
                  uint8_t target_type,  // BC, Group, Node
                  uint8_t target,       // target ID
                  uint8_t message_type) // type see CAN_MK.h
{
  uint32_t sID; //
  uint32_t tempID = source;
  tempID = tempID << 21; // shift sourceID to most left
  sID = tempID;          // and store
  tempID = target_type;  // next part target_type
  tempID = tempID << 19; // shift target_type behind sourceID
  sID = sID + tempID;    // add to sID
  tempID = message_type; // next part message_type
  tempID = tempID << 15; // shift message_type behind target_type
  sID = sID + tempID;    // add to sID
  tempID = target;       // next part target
  tempID = tempID << 7;  // shift targetID behind message_type
  sID = sID + tempID;    // add to sID
  return sID;            // return CAN ID
} // end encodeID

/* 
 * send a CAN message 
 * we get data len, target_type, targetID, message_type and optional 0 to 8 byte of data 
 * we fill the transmit buffer, add (our) sorceID, call the ID encode and send the message
 * count, destination_type, destination_ID, message_type
 * all 8 datab bytes are defined but sent are only the needed amount
 * as data not always needed these are optional predefined valuse
 */
void do_Send(uint8_t count, uint8_t dest, uint8_t destID, uint8_t type,
            uint8_t data = 0, uint8_t data_1 = 0, uint8_t data_2 = 0,
            uint8_t data_3 = 0, uint8_t data_4 = 0, uint8_t data_5 = 0,
            uint8_t data_6 = 0, uint8_t data_7 = 0)
{
  // fill the send buffer
  txBuf[M_Data] = data;                      // sender key
  txBuf[M_Data + 1] = data_1;                // local state
  txBuf[M_Data + 2] = data_2;                // local state
  txBuf[M_Data + 3] = data_3;                // local state
  txBuf[M_Data + 4] = data_4;                // local state
  txBuf[M_Data + 5] = data_5;                // local state
  txBuf[M_Data + 6] = data_6;                // local state
  txBuf[M_Data + 7] = data_7;                // local state
  txID = encodeID(myID, dest, destID, type); // encode the CAN ID
  CAN_Send(txID, count);                     // send it, count has the amount of data
} // END do_Send

/* 
 *  confBufShow
 *  we show EEPROM Config on the serial monitor if requested
 *  for debuging
 */
void confBufShow()
{
#ifdef SHOW_CONF
  uint16_t i;
  Serial.print(F("ConfSize "));
  Serial.println(sizeOfConf);
  for (i = 0; i < sizeOfConf; i++)
  {
    if ((i % 8) == 0)
    {
      Serial.println();
      Serial.print(i);
      if (i < 10)
      {
        Serial.print(F("   "));
      }
      else
      {
        if (i < 100)
        {
          Serial.print(F("  "));
        }
        else
        {
          Serial.print(F(" "));
        }
      }
    }
    toSerial(EEPROM.read(i));
  }
  Serial.println();
#endif
}
// END ConfBufShow

/* 
 *  Convert uint16_t to ASCII
 *  used to display time
 */
void intToAscii(uint16_t bin, char *timeStr)
{
  itoa(bin, timeStr, 10); // convert to ASCII string
  if (bin < 10)
  {
    timeStr[1] = timeStr[0];
    timeStr[0] = 0x30;
  }
}
// END intToAscii

/*
 * show time on LCD, we fill the line buffer at requested position
 * optional with seconds
 * expecting time[] = {h,m,s}
 */
void showTime(uint8_t col, const uint8_t *timeVal, uint8_t offset, bool second = false)
{
  char timeStr[3]; // reserve buffer
  // Hour
  intToAscii(timeVal[offset], timeStr); // convert
  if (timeStr[0] == 0x30)               // supress leeding "0"
  {
    lcdLine[col] = 0x20; // copy to LCD line buffer
  }
  else
  {
    lcdLine[col] = timeStr[0]; // copy to LCD line buffer
  }
  col++;                     // next position
  lcdLine[col] = timeStr[1]; // copy to LCD line buffer
  col++;                     // next position
  lcdLine[col] = ':';        // write delimiter
  col++;                     // next position
  // Minute
  intToAscii(timeVal[offset + 1], timeStr); // convert to ASCII string
  lcdLine[col] = timeStr[0];                // copy to LCD line buffer
  col++;                                    // next position
  lcdLine[col] = timeStr[1];                // copy to LCD line buffer
  // Second may be spuressed
  if (second)
  {
    col++;                                    // next position
    lcdLine[col] = ':';                       // write delimiter
    col++;                                    // next position
    intToAscii(timeVal[offset + 2], timeStr); // convert to ASCII string
    lcdLine[col] = timeStr[0];                // copy to LCD line buffer
    col++;                                    // next position
    lcdLine[col] = timeStr[1];                // copy to LCD line buffer
  }
}
//END showTime

/*
 * Setup we init everything
 */
void setup()
{
  Serial.begin(115200);       // Start serial monitor
  myID = (EEPROM.read(0));    // get myID from EEPROM to RAM
  myGroup = (EEPROM.read(1)); // get myGroup from EEPROM to RAM
#ifdef SHOW_CONF
  Serial.println(VERSION); //
  Serial.print("Node ID ");
  Serial.println(myID, HEX);
  Serial.print("Display addr: ");
  Serial.print(lcdAddr, HEX);
  Serial.print(" rows: ");
  Serial.print(lcdRows);
  Serial.print(" cols: ");
  Serial.println(lcdCols);
#endif
#ifdef I2C
  I2Cscan();
#endif

#ifdef SHOW_STATE
  uint8_t n = 0;                             // text index
  uint16_t i;                                // loop counter
  uint16_t s = expBase + (expBaseMult * 32); // expBase end
  for (i = expBase; i < s; i++)              // show state of expander ports
  {
    Serial.print(i); // show port number
    switch (n % 3)   // select text to print modulo 3
    {
    case 0:
      Serial.print(F(" state "));
      break;
    case 1:
      Serial.print(F(" source "));
      break;
    case 2:
      Serial.print(F(" key "));
    }
    toSerial(EEPROM.read(i)); // print value
    Serial.println();         // new line
    n++;                      // next text
  }
#endif

  sizeOfConf = confLen(); // caculate size of config data, store to ram
  confBufShow();          // show config to monitor on request
  // by default set SS pin to Output to disable it
  pinMode(SS, OUTPUT);
  digitalWrite(SS, LOW);
  if (sizeOfConf > 0)     // there may be problems with incompatible config
  {
    initLCD();            // based on config
    initGPIO();           // based on config
    showID(myID, 2000);   // optional time in ms
    startButtons();       // based on config
    initTimer();          // based on config
    initAddon();          // based on config
  }
  // everything is initialized, init CAN Bus now
  while (CAN_OK != CAN.begin(CAN_250KBPS, MCP_8MHz)) // init can bus : baudrate = 250K MCP 8MHz
  {
    delay(200);
  } // end while
  // show it on LCD
  if (configValue(displayBase, displayType) != 0xFF && lcdRows > 0)
  {
    lcd.setCursor(8, 2); // colume 0 row 2
    lcd.print(F("C "));
  }
} // end setup

/* Send a packet
 * we have extended frame id CanID contains 
 *  source, target type, target, message type
 * send the data:  id = 0x00, Extended Frame, data len = 8, txBuf: data txBuf
 * Extended Frame = 1.
 * CAN.sendMsgBuf(txID,1, 8, txBuf);    
 * More Info 
 * CAN.isRemoteRequest();  Check for Request
 * CAN.isExtendedFrame();  Check what Frame
 * return value is '0' for a negative response and '1' for a positive
 */
void CAN_Send(uint32_t txID, uint8_t len) // len = len of data
{
  CAN.sendMsgBuf(txID, 1, len, txBuf); // send an extended frame
} // end CAN_Send

/*
 * void decodeCanID(uint32_t rxID)
 * decodeCanID(uint32_t rxID)
 * decode CAN ID, results are going to rxHead 
 */
void decodeCanID(uint32_t rxID)
{
  // had problems defineing variables not at start
  uint32_t s_node = rxID;
  uint32_t t_type = rxID;
  uint32_t m_type = rxID;
  uint32_t t_node = rxID;
  // Source Node
  //        uint32_t s_node = rxID;
  s_node = s_node >> 21; // shift right, source ID remains
  rxHead[N_ID] = s_node; // store it
  s_node = s_node << 21; // shift left to get source ID part only

  // Target Type
  //        uint32_t t_type = rxID;
  t_type = t_type - s_node; // rxID - s_node leaves the rest
  t_type = t_type >> 19;    // shift right, target type remains
  rxHead[M_Dest] = t_type;  // store it
  t_type = t_type << 19;    // shift left to get target type only

  // Message Type
  //        uint32_t m_type = rxID;
  m_type = m_type - s_node; // rxID - s_node
  m_type = m_type - t_type; // and - t_type leaves the rest
  m_type = m_type >> 15;    // shift right message type remains
  rxHead[M_Type] = m_type;  // store it
  m_type = m_type << 15;    // shift left to get message type only

  // Target Node
  //        uint32_t t_node = rxID;
  t_node = t_node - s_node;   // rxID - s_node
  t_node = t_node - t_type;   // and - t_type
  t_node = t_node - m_type;   // and - m_type leaves the rest
  t_node = t_node >> 7;       // shift right t_node remains
  rxHead[M_Dest_ID] = t_node; // store it
}
// END decodeCanID

/*
 * loop here we do everything, we try to not use any delay()
 */
void loop()
{
  timer.update(); // check if we need it with TimeManager
  P1.tick();      // update pushbutton
  P2.tick();
  P3.tick();
  P4.tick();
  TimerManager::instance().update(); // update tiume manager
  checkReceive();                    // did we receceive a CAN message, needed things are done there
  checkSend();                       // do we have to dend data, needed things are done there
  checkPIR();                        // PIR not yet implemented
  portOff();                         // check if we have to switch ports of, depending on timers
                                     // should be called max twice per second
} // End LOOP

/*
 * hex2bcd(bin) returns a BCD coded value
 */
uint8_t hex2bcd(uint8_t x)
{
  unsigned char y;
  y = (x / 10) << 4;
  y = y | (x % 10);
  return (y);
} // end hex2bcd

/*
 * checkPIR() 
 * Checks the PIR of motiondetected send message dependig config
 * not shire about on t_ON and t_OFF, message say detected but relais 
 * works with t_Off
 */
void checkPIR()
{
  if (EEPROM.read(otherBase + otherPIR) != 0xFF) // PIR configured
  {
    pirValue = digitalRead(PIRpin);
    if (pirValue == HIGH)
    {
      if (pirState == LOW)
      {
        // GROUP NOT YET IMPLEMENTED
        do_Send(2, D_Node, EEPROM.read(otherBase + PIR_ID), Action,
               EEPROM.read(otherBase + PIR_port), t_OFF);
        // we have just turned on
        // We only want to print on the output change, not state
        pirState = HIGH;
        // Start LED
        // port uint8_t = ledPort(ledNum,ledCol)
        setLED(LED_1, gnLED, t_ON);
        setLED(LED_1, rdLED, t_OFF);
      }
    }
    else
    {
      if (pirState == HIGH)
      {
        do_Send(2, D_Node, EEPROM.read(otherBase + PIR_ID), Action,
               EEPROM.read(otherBase + PIR_port), t_ON);
        // We only want to print on the output change, not state
        pirState = LOW;
        // stop LED
        setLED(LED_1, gnLED, t_OFF);
        setLED(LED_1, rdLED, t_ON);
      }
    }
  }
} // end checkPIR

/* 
 * checkSend() 
 * send data if requested, if more than 8 byte transfer split in 8 bytes per message
 * it is secured by waiting on ACK for each message 
 * DATA SOURCE SHOULD BE VARIABLE IS FIXED EEPROM NOW
 */
void checkSend()
{
  // Sending Mass data
  if (dataTx == true) // data sending in progress
  {
    if (dataTxAck == true) // send next packed if any got Ack
    {
      dataTxAck = false;   // reset Ack
      if (dataTxCount > 0) // still data
      {
        uint8_t s_buffer[8];  // temp buffer
        uint8_t i;            // index to source
        uint8_t i_max;        // loop end
        uint8_t j;            // index to buffer
        if (dataTxCount >= 8) // more than 4 uint8_ts
        {
          i = dataTxIndex;       // set index to source
          i_max = i + 8;         // set index end
          j = 0;                 // init buffer index
          for (; i < i_max; i++) // loop read data
          {
            s_buffer[j] = EEPROM.read(i); // fill buffer from EEPROM
            dataTxCount--;                // dec to send
            j++;                          // inc buffer index
          }
          dataTxIndex = i; // save index to source
          delay(50);
          do_Send(8, D_Node, txDestination, Data,
                 s_buffer[0], s_buffer[1], s_buffer[2], s_buffer[3],
                 s_buffer[4], s_buffer[5], s_buffer[6], s_buffer[7]);
        }
        else // less than four uint8_t
        {
          i = dataTxIndex;            // set index to source
          i_max = i + dataTxCount;    // set index end to rest < 4
          uint8_t rest = dataTxCount; // save count
          j = 0;                      // init buffer index
          for (; i < i_max; i++)      // loop read data
          {
            s_buffer[j] = EEPROM.read(i); // fill buffer
            dataTxCount--;                // dec to send
            j++;                          // inc buffer index
          }
          dataTxIndex = i; // save index
          delay(50);       // smal delay should not harm
          // amount of data is dataTxCount (rest)
          do_Send(rest, D_Node, txDestination, Data,
                 s_buffer[0], s_buffer[1], s_buffer[2], s_buffer[3],
                 s_buffer[4], s_buffer[5], s_buffer[6], s_buffer[7]);
        }
      }    // end dataTxCount
      else // send finshed
      {
        dataTx = false;    // reset dataTx
        dataTxAck = false; // reset dataTxAck
        dataTxCount = 0;   // reset dataTxCount
        dataTxIndex = 0;   // resetdataTxIndex
        timer.stop();
      } // end if datasendCount
    }   // end if dataTxAck
  }     // end if dataTx
} // end checkSend

/* 
 * reset the node
 */
void (*resetFunc)(void) = 0; // declare reset function at address 0

/* 
 * If packet received check it and if for me do the requested 
 * fill rxHead with sender, target, target_ID, message_ID
 */
void checkReceive()
{

  if (CAN_MSGAVAIL == CAN.checkReceive()) // check if data coming
  {
    // clear the receive buffer
    for (uint8_t i = 0; i < 8; i++)
    {
      rxBuf[i] = 0;
    }
    CAN.readMsgBuf(&len, rxBuf);        // read data,  len: data length, rxBuf: data rxBuf
    rxID = CAN.getCanId();              // Get Sender ID
    // block bootloader message
    if (rxID == 0x1FFFFF01)
      return;
    if (rxID == 0x1FFFFF02)
      return;
    decodeCanID(rxID); // Fill rxHead
    if (configValue(displayBase, displayType) != 0xFF && lcdRows > 0)
    {
      lcd.setCursor(10, 2); // colume 0 row 2
      lcd.print(rxHead[N_ID], HEX);
      lcd.setCursor(12, 2); // colume 0 row 2
      lcd.print(rxHead[M_Dest_ID], HEX);
      lcd.setCursor(14, 2); // colume 0 row 2
      lcd.print(rxHead[M_Type], HEX);
      lcd.setCursor(19, 2); // colume 0 row 2
      inCount += 1;
      lcd.print(inCount % 10);
    }
#ifdef MONITOR
#ifdef NO_BC
    if (rxHead[M_Dest] != D_BC)
    {
#endif
      Serial.print(F("Length:")); // show data length
      Serial.println(len);
      Serial.print(F("rxID 0x")); // show sender canID
      Serial.println(rxID);
      Serial.print(F("my Node ID 0x")); // show myID
      Serial.print(myID, HEX);
      Serial.print(F(" Group 0x")); // show myGroup
      Serial.println(myGroup, HEX);
      Serial.print(F("got CAN Packet from 0x")); // show decoded sender
      Serial.println(rxHead[N_ID], HEX);
      Serial.print(F("DestType ")); // decoded target_type
      switch (rxHead[M_Dest])       // decode destination type
      {
      case D_BC:
        Serial.print(F("Broadcast "));
        break;
      case D_Group:
        Serial.print(F("Group "));
        break;
      case D_Node:
        Serial.print(F("Node "));
        break;
      default:
        Serial.print(F("UNKNOWN "));
        break;
      }
      Serial.print(F(" 0x")); // show target ID
      Serial.println(rxHead[M_Dest_ID], HEX);
      msgType(rxHead[M_Type]);  // decode message type
      uint8_t i;                // data start at rxBuf[0]
      for (i = 0; i < len; i++) // show data
      {
        Serial.print(F("Data+"));
        Serial.print(i);
        Serial.print(F(" 0x"));
        Serial.println(rxBuf[i], HEX);
      }
      Serial.println(F("==============="));
#ifdef NO_BC
    }
#endif
#endif
#ifdef LCDMON
    lcd.setCursor(0, 0);
    lcd.print(F("From 0x"));
    lcd.print(rxHead[N_ID], HEX);
    lcd.print(F(" To 0x"));
    lcd.print(rxHead[M_Dest], HEX);
    lcd.setCursor(0, 1);
    lcd.print(F("ID 0x"));
    lcd.print(rxHead[M_Dest_ID], HEX);
    lcd.print(F(" Type 0x"));
    lcd.print(rxHead[N_ID], HEX);
    lcd.setCursor(0, 2);
    for (i = 0; i < len; i++)
    {
      lcd.print(F("0x"));
      lcd.print(rxBuf[i], HEX);
      lcd.print(F(" "));
    }
#endif

    // checkreceive begins here
    switch (rxHead[M_Dest]) // who should get message
    {
    case D_BC:                // broadcast M_Dest_ID shoulbe always 0
      switch (rxHead[M_Type]) // Message Type
      {
      case C_Time:                         // time received, later only if configured
        showTime(0, rxBuf, M_Data, false); // at position TRUE shows seconds, fill only buffer
        lcdLineDisp(3);                    // needed do show define line line 4
        break;
      case S_Temp:       // temperature received, later only if configured
        showTH(9, 0x43); // at postion 8 C at end later fill only buffer
        lcdLineDisp(3);  // needed do show define line line 4
        break;
      case S_Humidity:    // Hhumidity received
        showTH(15, 0x25); // at posion 14 % at end fill only buffer
        lcdLineDisp(3);   // needed do show define line line 4
        break;
      }      // end M_TYPE while BC
      break; // case BC
    case D_Group:
      switch (rxHead[M_Type])
      {
      // GrFeedback is a feed to a group of buttons switching the same destination
      // if any of these buttons is hit all other correspondig LEDs should change to
      // the new state, rxBuf[0] = Data+0 has the group, rxBug[1] = Data+1 has the ne state
      // we loop over all configured buttons to see if we have a match with the group
      // if yes we set the LED to new state
      case GrFeedback:
        uint8_t s_Group = rxBuf[actPort];             // get the group
        uint16_t keyBase =  inputBase;
        for (uint8_t key = 0; key < 4; key++)   // loop over keys
        {
          uint16_t sourceBase = keyBase  + inputSkipPorts;
          for (uint8_t i = 0; i < 4; i++)       // loop over clicktype
          {
						// source group are in the range of 0x81 to 0xFE while target groups are from 0x01 to 0x7F
            if (configValue(sourceBase,inputSourceGr) == s_Group && s_Group > 0x80)
              {
                setLED(key,rdLED,rxBuf[1]);
                return;
								// !!!!!!!!!!!!!!!!!!!!
              }
            sourceBase = sourceBase + inputInnerLen;      // skip to next click type
          }
          keyBase += inputLen;
        }
        break;
      }
    case D_Node: // to Node direct
      // is it for me
      if (rxHead[M_Dest_ID] == myID) // it is for me
      {
        switch (rxHead[M_Type]) // Message Type
        {  //???????????????????????????
        case L_Info:   // we are announced getting a lot of data
        //Serial.println(F("*********** do_L_Info ***********"));
          do_L_Info(); // prepare receiving data, receive will be handled in checkReceive
          break; //?????????????????????
        case Ack: // we got an Ack, check if it is for an active transmission
          if (dataTx == true && rxHead[N_ID] == txDestination)
          {
            dataTxAck = true; // set if we are sending data and Ack comes from current destination
          }
          break;
        case Feedback:              // we got a feedback on request,
          setLED(rxBuf[M_Data], rdLED, rxBuf[M_Data + 1]);
          break;
        case Conf:                // it is a config request, RESET, get Config, send Config
          do_Conf(rxBuf[M_Data]); // do_Conf will do the rest
          break;
        case Action:   // it is an action requets, do something mostly set/unset an output
          do_Action(len); // do_Action will do the rest
          break;
        case Data:                                             // got a data packet
          do_Data();                                           // do_Data will do the rest
          break;                                               // end data
        case SetTime:                                          // we get new time to sync the RTC
          if (configValue(otherBase, otherRTC) == otherDS3231) // we have an RTC
          {                                                    // recreate the unix timestamp
            uint32_t unix_time;
            uint32_t tmp_time;
            tmp_time = rxBuf[M_Data];
            unix_time = tmp_time << 24;
            tmp_time = rxBuf[M_Data + 1];
            tmp_time = tmp_time << 16;
            unix_time = unix_time + tmp_time;
            tmp_time = rxBuf[M_Data + 2];
            tmp_time = tmp_time << 8;
            unix_time = unix_time + tmp_time;
            unix_time = unix_time + rxBuf[M_Data + 3];
            if (RTC.set(unix_time) == true)
            {
              sendAck(); // does not work, reason unknown, may be RTC.set does not resond proberly
            }
          }
          break;
        // whats the differnet between request, action and conf
        case Request: // we go a request check M_Data
          do_Request();
          break;
        default:
          Serial.println("UNKNOWN M_TYPE");
          break;
        } // end switch M_type
      }   // end if M_Dest_ID
      break;
      
    } // end switch M_Dest
  }   // end if MSGAVAIL
}
// END CheckReceive

/* 
 * send an ACK to sender 
 * zero data everything is in head (canID)
 */
void sendAck()
{
  do_Send(0, D_Node, rxHead[N_ID], Ack);
} // end sendAck

/* 
 * InitLCD()
 * inizialice LCD, clear line buffer 
 * we handle 2 and 4 lines and 16 and 20 colums
 */
void initLCD()
{
  uint8_t type = configValue(displayBase, displayType); // get the type
  switch (type)                                         // is a display configured
  {
  case displayLCD:                 // is there a display
    if (lcdCols > sizeof(lcdLine)) // check if buffer length
    {
      lcdCols = sizeof(lcdLine); // correct if buffer too small
    }
    lcd.init();      // initialize the lcd
    lcd.backlight(); // set baclight on
    lcdLineClear();  // clear the line buffer fill with spaces
    break;
  case displayGR: // graphic display NOT YET IMPLEMENTED
    break;
  case displayTouch: // touch display NOT YET IMPLEMENTED
    break;
  default:
    // no display
    break;
  }
} // END initLcd

/* 
 *  lcdLineClear()
 *  Clear lcdLine
*/
void lcdLineClear()
{
  uint8_t i;
  for (i = 0; i < lcdCols; i++) // Clear LCDLine
  {
    lcdLine[i] = ' ';
  }
} // end lcdLineClear

/* 
 *  lcdLineDisp()
 *  print line on display, we hand if request is for more lines than exist
 *  and if display is configured 
 *  graphic and touch display not recognized 
 */
void lcdLineDisp(uint8_t line)
{
  // check for Display
  if (configValue(displayBase, displayType) != 0xFF)
  {
    switch (lcdRows)
    {
    case 2:
      if (line >= 2)
      {
        line = line - 2;
      }
      break;
    default:
      // no clear, buffer should becleared
      lcd.setCursor(0, line);
      for (uint8_t i = 0; i < lcdCols; i++) // len should depend from display
      {
        lcd.write(lcdLine[i]);
      } // end for
      break;
    } // end switch
  }
} // end lcdLineDisp

/* 
 *  showTH()
 *  Build temperature/huminity in display buffer
 */
void showTH(uint8_t col, unsigned char einheit)
{
  uint8_t i;
  uint8_t stopit = col + 4; // 2 digits, dot, digit
  uint8_t offset = M_Data;
  for (i = col; i < stopit; i++)
  {
    lcdLine[i] = rxBuf[offset];
    offset++;
  }                     // end for
  lcdLine[i] = einheit; // "C", "%"
} // end showTH

/*
 * void initPort(uint8_t port)
 * init a led port, save program space
 */
void initPort(uint8_t port)
{
  pinMode(port, OUTPUT);   // setup port
  digitalWrite(port, LOW); // init port
}
// END initPort

/* 
 * void initGPIO()
 * Initialize GPIO mostly outputs, push button port will be inizialzed by library
 * we init the LED ports if defined in config
 * set configured LED ports to OUTPUT, LEDs itself are set by showID
 */
void initGPIO()
{
  uint8_t port;
  for (uint8_t i = 0; i < 12; i++)
  {
    if (i < 4)
    {
      port = isLED(i, gnLED);
    }
    else
    {
      if (i < 8)
      {
        port = isLED(i, rdLED);
      }
      else
      {
        port = isLED(i, yeLED);
      }
    }
    if (port < 0xFF)
    {
        pinMode(port,OUTPUT);
    }
  }
// END iniGPIO

  // show init state on LCD
  if (configValue(displayBase, displayType) != 0xFF && lcdRows > 0)
  {
    lcd.setCursor(0, 2); // colume 0 row 2
    lcd.print(F("G "));
  }
}
// END initGPIO

/*
 * startButtons()
 * Initialize the button functions
 * Start objects
 * assign functions
 * button#.attachClick(function);
 * button#.attachDoubleClick(function);
 * button#.attachLongPressStart(function);
 * button#.attachLongPressStop(function);
 * button#.attachDuringLongPress(function);
 */
void startButtons()
{
  uint8_t buttonLoop;   // loop about all buttons
  uint8_t functionLoop; // inner loop
  uint16_t b_offset = inputBase;      // in EEPROM
  uint16_t suboffset;   // in EEPROM
  if (inputNum > 0)     // we have buttons/switches defined in CAN_MK.h
  {
    // we loop about all available buttons
    for (buttonLoop = 0; buttonLoop < inputNum; buttonLoop++)
    {
      switch (configValue(b_offset,inputType))  // type of button
      {
        // which button type
        case pushButton: revPushButton:
          suboffset = b_offset + inputSkipPorts;    // start off inner part
          // we loop over the innerpart, possible click types
          for (functionLoop = 0; functionLoop < maxClick; functionLoop++)
          {
            // which click type
            // we may set ClickTicks, PressTicks, DebounceTicks
            switch (configValue(suboffset,0))
            {
              case nClick:                      // normal click
                // init nClick service functions
                if (buttonLoop == 0)            // button 1
                  P1.attachClick(onP1Click);
                if (buttonLoop == 1)            // button 2
                  P2.attachClick(onP2Click);
                if (buttonLoop == 2)            // button 3
                  P3.attachClick(onP3Click);
                if (buttonLoop == 3)            // button 4
                  P4.attachClick(onP4Click);
                break;
              case dClick:                      // double Click
                // init dClick service functions
                if (buttonLoop == 0)            // button 1
                  P1.attachDoubleClick(onP1Double);
                if (buttonLoop == 1)            // button 2
                  P2.attachDoubleClick(onP2Double);
                if (buttonLoop == 2)            // button 3
                  P3.attachDoubleClick(onP3Double);
                if (buttonLoop == 3)            // button 4
                  P4.attachDoubleClick(onP4Double);
                break;
              case wlClick:                     // while Click, not yet implemented
                break;
              case elClick:                     // end long Click
                // init elClicl service functions
                if (buttonLoop == 0)            // button 1
                  P1.attachLongPressStop(onP1Long);
                if (buttonLoop == 1)            // button 2
                  P2.attachLongPressStop(onP2Long);
                if (buttonLoop == 2)            // button 3
                  P3.attachLongPressStop(onP3Long);
                if (buttonLoop == 3)            // button 4
                  P4.attachLongPressStop(onP4Long);
                break;
              default:
                break;
            } // end click type
            // calc offset to next type
            suboffset = suboffset + inputInnerLen;   // skip target port etc.
          }
          break;
        default:                                // no other types implemeted
          break;
      }
      // cal offset to next button
      b_offset = b_offset + inputLen;
    }
  }
  // show buttons initialized on LCD
  if (configValue(displayBase, displayType) != 0xFF && lcdRows > 0)
  {
    lcd.setCursor(2, 2); // colume 0 row 2
    lcd.print(F("B "));
  }
}
// END startButtons

/*
 * onP1Click()
 * Callback function P1
 */
void onP1Click()
{
  uint16_t base = inBase(0, nClick); // calc base inbase(button #, clickType)
  sendKey(base, 0);                  // send packet dependig config
} // end onP1Click

/*
 * onP1Long()
 */
void onP1Long()
{
  uint16_t base = inBase(0, elClick); // calc base inbase(button #, clickType)
  sendKey(base, 0);                   // send packet dependig config
} // end onP1Long

/*
 * onP1Double()
 */
void onP1Double()
{
  uint16_t base = inBase(0, dClick); // calc base inbase(button #, clickType)
  sendKey(base, 0);                  // send packet dependig config
} // end onP1Double

/*
 * onP2Click()
 * Callback function P2
 */
void onP2Click()
{
  uint16_t base = inBase(1, nClick); // calc base inbase(button #, clickType)
  sendKey(base, 1);                  // send packet dependig config
} // end onP2Click

/*
 * onP2Long()
 */
void onP2Long()
{
  uint16_t base = inBase(1, elClick); // calc base inbase(button #, clickType)
  sendKey(base, 1);                   // send packet dependig config
} // onP2Long

/*
 * onP2Double()
 */
void onP2Double()
{
  uint16_t base = inBase(1, dClick); // calc base inbase(button #, clickType)
  sendKey(base, 1);                  // send packet dependig config
} // end onP2Double

/*
 * onP3Click()
 * Callback function P3
 */

void onP3Click()
{
  uint16_t base = inBase(2, nClick); // calc base inbase(button #, clickType)
  sendKey(base, 2);                  // send packet dependig config
} //end onP3Click

/*
 * onP3Long()
 */
void onP3Long()
{
  uint16_t base = inBase(2, elClick); // calc base inbase(button #, clickType)
  sendKey(base, 2);                   // send packet dependig config
} // end onP3Long

/*
 * onP3Double()
 */
void onP3Double()
{
  uint16_t base = inBase(2, dClick); // calc base inbase(button #, clickType)
  sendKey(base, 2);                  // send packet dependig config
} // end onP3Double

/*
 * onP4Click()
 * Callback function P4
 */
void onP4Click()
{
  uint16_t base = inBase(3, nClick); // calc base inbase(button #, clickType)
  sendKey(base, 3);                  // send packet dependig config
} // end onP4Click

/*
 * onP4Long()
 */
void onP4Long()
{
  uint16_t base = inBase(3, elClick); // calc base inbase(button #, clickType)
  sendKey(base, 3);                   // send packet dependig config
} // end onP4Long

/*
 * onP4Double()
 */
void onP4Double()
{
  uint16_t base = inBase(3, dClick); // calc base inbase(button #, clickType)
  sendKey(base, 3);                  // send packet dependig config
} // end onP4Double

/*
 * dataTxStop()
 * we tried to send data, but there was a timeout 
 * reset sending and show information depending on config
 */
void dataTxStop()
{
  dataTx = false; // reset  ?? more
  showError(E_TimeoutSend);
} // end data send stop

/*
 * dataRxStop()
 * we had a request to get data, but there was a timeout
 * reset receiving and show information depending on config
 */
void dataRxStop()
{
  dataRx = false; // reset ?? more
  showError(E_TimeoutReceive);
} // end dataRxStop

/*
 * void showError(uint16_t code)
 * show Error if any dependig config
 */
void showError(uint16_t code)
{
  lcdLineClear();
  switch (code)
  {
  case E_TimeoutSend:
    Serial.println(Timeout_Send);
    break;
  case E_TimeoutReceive:
    Serial.println(Timeout_Receive);
    break;
  }
} // show error

/*
 * void initTimer()
 * initialize timers for different tasks
 * we should think to move to long timers
 * we do not have an RTC on every node so we have no 1 second interrupt
 * but we can generate a 1 second interval with timer
 */
void initTimer()
{
  switch (configValue(sensorBase, sensorTempType))
  {
  case sensorDHT:
    DHTtemp.setInterval(61000); // broadcast temperature every 6,1 second
    DHTtemp.setCallback(bcTemp);
    break;
  }
  switch (configValue(sensorBase, sensorHumitType))
  {
  case sensorDHThum:
    DHThum.setInterval(124000); // broadcast humidity every 12,4 seconds
    DHThum.setCallback(bcHum);
    break;
  }
  switch (configValue(otherBase, otherRTC)) // broadcast time every 10 seconds
  {
  case otherDS3231:
    RTCtime.setInterval(10000);
    RTCtime.setCallback(bcTime);
    break;
  }
  TimerManager::instance().start();
  if (configValue(displayBase, displayType) != 0xFF && lcdRows > 0)
  {
    lcd.setCursor(4, 2); // colume 0 row 2
    lcd.print(F("T "));
  }
}
// END initTimer

/*
 * bcTemp()
 * broadcast temperature
 */
void bcTemp()
{
  float h = dht.readHumidity();
  float t = dht.readTemperature(); // option TRUE for Fahrenheit
  if (isnan(h) || isnan(t))
  {
    Serial.println(DHT_FAIL);
  }
  else
  {
    char result[4];
    dtostrf(t, 2, 1, result);
    // void do_Send(uint8_t count,uint8_t dest, uint8_t destID, uint8_t  type,
    do_Send(5, D_BC, 0, Info, S_Temp, result[0], result[1], result[2], result[3]);
  } // values avail
} // end bcTemp

/*
 * bcHum()
 * broadcast humidity
 */
void bcHum()
{
  float h = dht.readHumidity();
  float t = dht.readTemperature(); // option TRUE for Fahrenheit
  if (isnan(h) || isnan(t))
  {
    Serial.println(DHT_FAIL);
  }
  else
  {
    char result[4];
    dtostrf(h, 2, 1, result);
    do_Send(5, D_BC, 0, Info, S_Humidity, result[0], result[1], result[2], result[3]);
  } // end valus avail
} // end bcHum

/*
 * bcTime()
 * broadcast time
 */
void bcTime()
{
  uint8_t timeVal[3];
  time_t time = RTC.get();
  timeVal[0] = hour(time);
  timeVal[0]++; // convert to local time !! SummerTime !!
  if (timeVal[0] == 24)
  {
    timeVal[0] = 0;
  }
  timeVal[1] = minute(time);
  timeVal[2] = second(time);
  do_Send(4, D_BC, 0, Info, C_Time, timeVal[0], timeVal[1], timeVal[2]);
  // display time o local LCD
  uint8_t type = configValue(displayBase, displayType);
  switch (type) // is a display configured
  {
  case displayLCD:
    showTime(12, timeVal, 0, true);
    lcdLineDisp(3);
    break;
  } // end bcTime
}

/*
 * void initAddon()
 * init peripherials RTC, DHT etc
 * to switch output use expanderx.digitalWrite(n, HIGH/LOW);
 */
void initAddon()
{
  uint8_t i;             // expander loop
  uint8_t j;             // inner expander loop
  uint8_t isDHT = false; // default no
  uint8_t isBME = false; // default no
  uint8_t isDS = false;  // default no
  uint8_t isState;
  // check for sound
  if (configValue(piepsBase, piepsPort) != 0xFF)
  {
    switch (configValue(piepsBase, piepsType))
    {
    case piepsPassiv:
      break; //to be implemented
    case piepsActiv:
      break; //to be implemented
    }
  }
  // check for configured temperatur sensor
  switch (configValue(sensorBase, sensorTempType))
  {
  case sensorDHT: // there is a DHT 22 temperature and humidity
    dht.begin();
    isDHT = true; // to be implemented
    break;
  case sensorDS18B20: // there is a DS18B20 temperarture
    isDS = true;      // to be implemented
    break;
  case sensorBME280: // there is a BME280 temperature, humidity and air pressure
    isBME = true;    // to be implemented
    break;
  }
  // check for configured humidity sensor
  switch (configValue(sensorBase, sensorHumitType))
  {
  case sensorDHThum: // there is a DHT 22
    if (!isDHT)      // if not yet inititialized
    {
      dht.begin(); // init it
      isDHT = true;
    }
    break;
  case sensorBME280hum: // there is a BME280
    if (!isBME)         // if not yet inititialized
    {
      isBME = true;
    }
    break;
  }
  // check for air pressure sensor
  switch (configValue(sensorBase, sensorPressureType))
  {
  case sensorBME280pres: // ther is a BME280
    if (!isBME)          // if not yet inititialized
    {
      isBME = true;
    }
    break;
  }
  // ckeck for other peripherals
  switch (configValue(otherBase, otherRTC))
  {
  case otherDS3231:
    pinMode(RTC_1HZ_PIN, INPUT_PULLUP); // enable pullup on uint16_terrupt pin (RTC SQW pin is open drain)
    attachInterrupt(digitalPinToInterrupt(RTC_1HZ_PIN), myTimers, FALLING);
    RTC.squareWave(SQWAVE_1_HZ); // 1 Hz square wave
    // Initialize EEPROM library.
    // eeprom.initialize();
    break;
  }
  switch (configValue(otherBase, otherIR)) // there is a IR receiver
  {
  case IRdiode:
    break;
  }
  switch (configValue(otherBase, otherPIR)) // there is a PIR
  {
  case normalPIR:
    pinMode(PIRpin, INPUT_PULLUP);
    break;
  }
  // there may be a radar sensor !!
  // init expander ports
  for (i = 0; i < 4; i++)
  {
    switch (i)
    {
    case 0:
      if (configValue(otherBase, otherExtendAdr1) != 0xFF) // exists if has valid address
      {
        expander1.begin(configValue(otherBase, otherExtendAdr1)); // init it
        for (j = 0; j < 8; j++)
        {
          expander1.pinMode(j, OUTPUT);     // set port as OUTPUT
          isState = getExpState(j, 0);      // get last state
          relais_1_Do(j, isState);          // set output to state power fail
        }
      }
      break;
    case 1:
      if (configValue(otherBase, otherExtendAdr2) != 0xFF)
      {
        expander2.begin(configValue(otherBase, otherExtendAdr2));
        for (j = 0; j < 8; j++)
        {
          expander2.pinMode(j, OUTPUT);
          isState = getExpState(j, 8);
          relais_2_Do(j, isState);            // set output to state power fail
        }
      }
      break;
    case 2:
      if (configValue(otherBase, otherExtendAdr3) != 0xFF)
      {
        expander3.begin(configValue(otherBase, otherExtendAdr3));
        for (j = 0; j < 8; j++)
        {
          expander3.pinMode(j, OUTPUT);
          isState = getExpState(j, 16);
          relais_3_Do(j, isState);            // set output to state power fail
        }
      }
      break;
    case 3:
      if (configValue(otherBase, otherExtendAdr4) != 0xFF)
      {
        expander4.begin(configValue(otherBase, otherExtendAdr4));
        for (j = 0; j < 8; j++)
        {
          expander4.pinMode(j, OUTPUT);
          isState = getExpState(j, 24);
          relais_4_Do(j, isState);            // set output to state power fail
        }
      }
      break;
    }
  }
  // show init on LCD
  if (configValue(displayBase, displayType) != 0xFF && lcdRows > 0)
  {
    lcd.setCursor(6, 2);                      // colume 0 row 2
    lcd.print(F("A "));
  }
}
// END initAddon

/*
 * uint8_t getExpState(i,offset)
 * get the last state from EEPROM
 * return state
 */
uint8_t getExpState(uint8_t pin, uint8_t relaisSet)
{
  uint8_t portState;
  pin = pin + relaisSet; // we calulate the port number
  portState = configValue(expBase, (pin * expBaseMult));
  // correct if undefined value
  if (portState > 0x01) // undefined, set it to 0x00 = OFF
  {
    portState = 0x00;
  }
  return portState;
}
// END getExpState

/*
 * void relais_n_Do(uint8_t pin)
 * relais block n
 * Set output
 * we have to XOR portset for relais
 * so we can pass action direct
 * Check port existance
 */
void relais_1_Do(uint8_t pin, uint8_t portset)
{
  if (configValue(otherBase, otherExtendAdr1) != 0xFF)
  {
    expander1.digitalWrite(pin, portset ^ 0x01); // set port
    fillLCDline(1, portset, 10 + pin);           // display state on LCD
  }
}
// END relais_1_DO

void relais_2_Do(uint8_t pin, uint8_t portset)
{
  if (configValue(otherBase, otherExtendAdr2) != 0xFF)
  {
    expander2.digitalWrite(pin, portset ^ 0x01); // set port
    fillLCDline(1, portset, pin);
  }
}
// END relais_2_DO

void relais_3_Do(uint8_t pin, uint8_t portset)
{
  if (configValue(otherBase, otherExtendAdr3) != 0xFF)
  {
    expander3.digitalWrite(pin, portset ^ 0x01); // set port
    fillLCDline(0, portset, 10 + pin);
  }
}
// END relais_3_DO

void relais_4_Do(uint8_t pin, uint8_t portset)
{
  if (configValue(otherBase, otherExtendAdr4) != 0xFF)
  {
    expander4.digitalWrite(pin, portset ^ 0x01); // set port
    fillLCDline(0, portset, pin);
  }
}
// END relais_4_DO

/*
 * uint8_t modAction(uint16_t offset, uint8_t action)
 * EEPROM helds actual state
 * correct action if toggle and set action for timed
 * remember new state in EEPROM
 * correct portset, we will use return value for write to expanders
 */
uint8_t modAction(uint16_t offset, uint8_t action)
{
  uint8_t portset;
  switch (action)
  {
  case t_Toggle:                     // if action we have to check actual state
    if (EEPROM.read(offset) == 0x00) // state is OFF
    {
      EEPROM.write(offset, 0x01);
      portset = 0x01; // we have to turn ON
    }
    else
    {
      EEPROM.write(offset, 0x00);
      portset = 0x00; // we have to turn OFF
    }
    break;
  case t_Timed:
    EEPROM.write(offset, 0x01);
    portset = 0x01; // lets turn on
    break;
  default:
    EEPROM.write(offset, action);
    portset = action; // return just action
    break;
  }
  return portset;
} // end modAction

/*
 * void msgType(uint8_t type)
 * decode message type for monitoring on request
 */
void msgType(uint8_t type)
{
#ifdef MONITOR
  Serial.print(F("Message Type "));
  switch (type)
  {
  case Request:
    Serial.println(F("Request"));
    switch (rxBuf[M_Data])
    {
    case PingRequest:
      Serial.println(F("Ping Request"));
      break;
    case PortState:
      Serial.println(F("Port State"));
      break;
    default:
      Serial.print(rxBuf[M_Data], HEX);
      Serial.println(" UNKNOWN");
      break;
    }
    break;
  case PingReply:
    Serial.println(F("Ping Reply"));
    break;
  case Action:
    Serial.println(F("Action"));
    break;
  case U_Action:
    Serial.println(F("U_Action"));
    break;
  case L_Info:
    Serial.println(F("L_Info"));
    break;
  case Feedback:
    Serial.println(F("isFeedBack"));
    break;
  case GrFeedback:
    Serial.println(F("grFeedBack"));
    break;
  case Ack:
    Serial.println(F("Ack"));
    break;
  case Data:
    Serial.println(F("Data"));
    break;
  case Conf:
    Serial.println(F("Conf"));
    switch (rxBuf[M_Data])
    {
    case ConfReset:
      Serial.println(F("reset"));
      break;
    case ConfRequest:
      Serial.println(F("query"));
      break;
    case ConfNewID:
      Serial.println(F("new ID"));
      break;
    case ConfNewGr:
      Serial.println(F("new Gr"));
      break;
    case ConfNew:
      Serial.println(F("new Conf"));
      break;
    default:
      Serial.print(rxBuf[M_Data], HEX);
      Serial.println(" UNKNOWN");
    }
    break;
  case SetTime:
    Serial.println(F("SetTime"));
    break;
  case Info:
    Serial.print(F("Info "));
    switch (rxBuf[M_Data])
    {
    case C_Time:
      Serial.println(F(" local Time"));
      break;
    case C_G_Time:
      Serial.println(F(" World Time"));
      break;
    case C_Sunraise:
      Serial.println(F(" Sunraise"));
      break;
    case C_Sunset:
      Serial.println(F(" Sunset"));
      break;
    case S_Temp:
      Serial.println(F(" Temperature"));
      break;
    case S_Pressure:
      Serial.println(F(" Pressure"));
      break;
    case S_Humidity:
      Serial.println(F(" Humidity"));
      break;
    default:
      Serial.print(type, HEX);
      Serial.println(F(" UNKNOWN"));
      break;
    }
    break;
  default:
    Serial.print(type, HEX);
    Serial.println(F(" UNKNOWN"));
    break;
  }
#endif
} // end msgType

/*
 * uint16_t configValue(uint16_t base, uint16_t offset)
 * simplify reading from EEPROM
 * we get the base pointer to the to the specific table
 * and the offset within the table
 * we return the correspondig value
 * get value from configuration, read from EEPROM
 */
uint16_t configValue(uint16_t base, uint16_t offset)
{
  uint16_t erg = EEPROM.read(base + offset);
  return erg;
} // end configValue

/*
 * toSerial(uint16_t value)
 * make hex output nice
 */
void toSerial(uint16_t value)
{
  if (value > 15)
  {
    Serial.print(F("0x"));
  }
  else
  {
    Serial.print(F("0x0"));
  }
  Serial.print(value, HEX);
  Serial.print(F(" "));
} // end toSerial

/*
 * inBase(uint8_t buttonNum ,uint8_t clickType)
 * calculate base for button
 */
uint16_t inBase(uint8_t buttonNum, uint8_t clickType) // calc base inbase(button #, clickType)
{
  uint16_t base;
  base = inputBase + (buttonNum * inputLen); // taster (ab 0)  * inputlen (element len)
  base += inputSkipPorts;                    // skip ports
  base = base + (clickType * inputInnerLen); // base
  return base;
} // end inBase

/*
 * sendKey(uint16_t base, uint8_t key)
 * send key (pushed button, switched key) depending config
 * request an action for port on target
 * base has offset to inputInnerLen of key clickmode selected
 * timed not working with group and sourcegroup 
 */
void sendKey(uint16_t base, uint8_t key)
{
  uint8_t len = 4;                     									// data len  4 default
  uint8_t reqFeedBack;																	// has geddback type
  uint8_t t_Func = configValue(base, inputTargetFunc); 	// get the function t_ON, t_OFF ..
	// handle feedback request	
  uint8_t option = configValue(base, inputSourceGr);		// get a possible source group
  // handle feedback
  if (option < 0xFF)                    								// feedback to group
  {
    reqFeedBack = grFeedBackReq;												// request a group feedback
    len = 5;																						// source group exists adjust data length
		do_Send(len, D_Node, configValue(base, inputTargetID), Action,
					configValue(base, inputTargetPort), 					// data 0
					t_Func,                             					// data 1
					reqFeedBack,                        					// data 2
					key,                                					// data 3
					option);                            					// data 4 option
		return;		
		// !!!!!!!!!!!!!!!!!!! 
  }
  else                                   								// default feedback request 
  {
    reqFeedBack = feedBackReq;           								// we want a feedback always
  }
  // handle a target group
	// we reuse otion
  option = configValue(base, inputTargetGr); 						// get a possible target goup
  if (option < 0xFF)
  {
    len = 5;                                    				// atarget group exists adjust data length
  }
  do_Send(len, D_Node, configValue(base, inputTargetID), Action,
         configValue(base, inputTargetPort), // data 0
         t_Func,                             // data 1
         reqFeedBack,                        // data 2
         key,                                // data 3
         option);                            // data 4 option
}
// END sendKey

/*
 * void setLED(uint8_t led, uint8_t color, uint8_t onOFF)
 * switch a LED beside a key (button, switch) on, off, toogle or dimm
 * button hardware has max four buttons with four two color LEDs
 * key numbers are 0,1,2,3, colors are rdLED, gnLED, onOFF 1,0
 */
void setLED(uint8_t led, uint8_t color, uint8_t onOFF)
{
  uint8_t port = isLED(led,color);      // get the correspondig port from config
  if (port < 0xFF)                      // enabled
  {
    uint8_t value = isDim(led,color);   // get the dim value
    switch (onOFF)
    {
    case 0: // switch it off
      writeLED(port, value);          // set to default 
      break;
    case 1: // switch on
      writeLED(port, 0xFF);
      break;
    case 2: // toggle
      if (digitalRead(port) == HIGH)  // LED is full on
      {
        writeLED(port, value);        // set to default
      }
      else
      {
        writeLED(port, 0xFF);         // set LED full on   
      }
      break;
    }
  }
}
// END setLED

/*
 * void writeLED(uint8_t port, uint8_t value)
 * set LED off, on, dimmed
 */
void writeLED(uint8_t port, uint8_t value)
{
  if (value == 0x00 || value == 0xFF)
  {
    digitalWrite(port,value);
  }
  else
  {
    analogWrite(port, value);
  }
} // end writeLED

/*
 * do_L_Info()
 * we got L_Info length information about receiving data
 * initialze data receive
 */
void do_L_Info()
{
  dataRxType = rxBuf[M_Data]; // type of recive data
  // dataRxType is not needed elsewhere
  dataRxSubType = rxBuf[M_Data + 1]; //
  dataRxCount = rxBuf[M_Data + 2];   // amount data
  dataRxStore = dataRxCount;         // we need it later
  dataRxIndex = 0;
  dataRx = true; // let us wait for data
  timer.setTimeout(dataRxTimeout);
  timer.setCallback(dataRxStop);
  timer.start();
  if (dataRxSubType == ConfNew || dataRxCount <= sizeof(dataRxBuffer))
  {
    sendAck();
  }
} // end do_L_Info

/*
 * void do_Data()
 * we received a data packet
 * only for new conf till now
 */
void do_Data()
{
  if (dataRx == true) // why do we check that here, can we come here without dataRX equals true
  {
    uint16_t index = dataRxIndex; // last position in target
    uint8_t k = M_Data;           // index in message data part
    uint8_t j = 8;                // loop end 8 bytes or less
    if (dataRxCount < 8)          // if rest less than 8 byte
    {
      j = dataRxCount; // set new loop end if less than 8 byte received
    }
    switch (dataRxSubType) // check for data type
    {
    case ConfNew: // we get a new conf
      for (uint8_t i = 0; i < j; i++)
      {
        if (index > 1) // we skip nodeID and group
        {
          EEPROM.update(index, rxBuf[k]); // store to EEPROM
        }
        k++;           // inc pointer to receive buffer
        index++;       // inc pointer to target
        dataRxCount--; // dec received counter
      }
      break;
    }
    sendAck();            // ack the message
    dataRxIndex = index;  // save next target position
    if (dataRxCount == 0) // last byte received
    {
      dataRx = false;    // reset flag
      dataRxSubType = 0; // reset info
      timer.stop();      // stop timeout check
      // if we got other data than config
      // we have to implement it here
    }
  } // end if receive = true
}
// END do_Data

/*
 * void do_Conf(uint8_t request)
 * we have a configuration request
 */
void do_Conf(uint8_t request)
{
  switch (request) // Conf
  {
  case ConfReset:
    if (rxBuf[M_Data + 1] > 0) // we have a delay power on reset
    {
      if (isLED(LED_1, rdLED) < 0xFF)
      {
        digitalWrite(isLED(LED_1, rdLED), HIGH);
      }
      delay(rxBuf[M_Data + 1] * 1000); // delay request seconds
      wdt_enable(WDTO_15MS);           // provides a Soft Reset when connected to FDTI Port, that provides power
      for (;;)
      {
      } // wait for it to reset and start over at the boot loader...
    }
    // RESET NODE
    resetFunc(); // call reset
    // ---- PROGRAM ENDS HERE ______
    break; //ConfReset
  // got New ID
  case ConfNewID:
    EEPROM.update(N_ID, rxBuf[M_Data + 1]); // M_Data has request SubType
    break;                                  // confNewID
  // we shall send our actual config
  case ConfRequest:               // initalize data send
    dataTxCount = sizeOfConf;     // amount of data to send
    txDestination = rxHead[N_ID]; // sender of request

    // NOTE WE ASSUME CONFIGURATION DOES NOT GROW ABOVE 256 BYTES

    do_Send(1, D_Node, txDestination, L_Info, sizeOfConf);
    dataTx = true;
    dataTxIndex = 0;
    timer.setTimeout(dataTxTimeout);
    timer.setCallback(dataTxStop);
    timer.start();
    break;
  } // end switch
}
// what about getting a new conf??
// END do_Conf

/*
 * void do_Action()
 * we have an action request to set an output, actually extender port
 * switch a port on, check for portextender existance later
 * we store the actual state in any case
 * we get a port number and have to translate it to pin number of a 
 * corresponding port extender
 */
void do_Action(uint8_t len)
{
  uint8_t portset;                                  // how to set port
  uint8_t port = rxBuf[actPort];                    // get port (M_Data)
  uint8_t action = rxBuf[actAct];                   // action   (M_Data+1)
	uint8_t feedback = rxBuf[actFeedBack];						// feedback	(M_Data+2)
  uint8_t key = rxBuf[actKey];                      // key   		(M_Data+3)
	uint8_t time_tgroup = rxBuf[actTime];							// option time/t_group (M_Data+4)

  uint16_t offset = expBase + (expBaseMult * port); // correct offset
 
  // remember state of output in EEPROM
  // modAction toogles state in EEPROM if togglge, set state to ON in EEPROM if timed
  // returns the correct value to set the expander 
  portset = modAction(offset, action);              // returns 0x00 for OFF 0x01 for ON
  
  // portset has new state, from toogle or timed or just on/off from action
  // we have to set the output now
	
	// target groups go from 0x01 to 0x7F while source groups go from 0x81 to 0xFE
	if (action != t_Timed && len > 4 && time_tgroup < 0x80)		// we have a target group request
	{
		// search correspondig target group members
		for (uint8_t i = 0;i < 32;i++)
		{
			if (configValue(grBase,i) == time_tgroup)
			{
				setPort(i,portset);									// we found a target group
			}
		}
	}
	else
	{
		setPort(port,portset);									// normal action
	}
  
	// if timed request we must start timer to switch off port automatic

  if (action == t_Timed) 							// activate timer
  {
    longTimers[port] = time_tgroup; 					// store time (M_Data+4)
  }
  // we do a feedback by default
  //  rxBuf[actFeedBack] = feedBackReq;
  switch (feedback) // if feedback request (M_Data+2)
  {
  case feedBackReq: // it is not M_Type but request
    // do_Send(2,D_Node,rxHead[N_ID],Feedback,key,isState);
    // len, Target_Type, Target_ID,Message_Type,key,state
    do_Send(2, D_Node, rxHead[N_ID], Feedback, key, portset);
    if (action == t_Timed)
    {
      EEPROM.write(offset + expKey,key);									// save key for feedback
			EEPROM.write(offset + expSource,rxHead[N_ID]);			// save source for feedback
    }
    break;
  case grFeedBackReq: // it is not M_Type but request
    do_Send(2,D_Group,0,GrFeedback,rxBuf[4],portset);
    break;
  default:
    break;
  }
}
// END do_Action

/*
 * void myTimers()
 * interrupt routine second timer
 * if we have an RTC (we must have at least one in the system
 * best where the relais a switched on and of
 * we get a 1 second interval squarewave triggering the interrupt 
 * so we have timers with a multiple of 1 second 255 seconds max 
 * we initialize 32 for 32 output
 * we use them to reset timed outputs
 * to shorten the interrupt routine we just set
 * flag for what to do on timeout
 */
//
void myTimers()
{
  uint8_t test;
  for (uint8_t i = 0; i < 32; i++) // loop about max posible ports
  {
    test = longTimers[i]; // dec timecounter if activ (value > 0)
    if (test > 0)         // timer active
    {
      test--;               // dec
      longTimers[i] = test; // set finished flag if reached zero
      if (test == 0)        // timeout
      {
        longState[i] = 1; // set flag, loop will call do the work
      }
    }
  }
}
// END myTimers

/*
 * void portOff()
 * called from loop 
 * switch a port off if flag set, reset flag
 * longState array is set by timer
 * we have to do the rest expand existance check later
 * no group support
 */
void portOff()
{
  uint8_t pin = 0xFF;              // undefined pin
  uint8_t source;                   // undefined source
  uint8_t key;                      // undefined key
  for (uint8_t i = 0; i < 32; i++)  // loop about state array
  {
    if (longState[i] == 1) // check if timed port is on
    {
      longState[i] = 0;                               // reset
      uint16_t index = expBase + (i * expBaseMult);   // base index
      EEPROM.write(index + expState, 0);              // reset state
      source = configValue(index,expSource);
      if (source < 0xFF) // check if to send feedback
      {
        key = configValue(index,expKey); // get the stored key
        do_Send(2, D_Node, source, Feedback, key, 0);   // send feedback state Off
        EEPROM.write(index + expSource, 0xFF);          // reset target
        EEPROM.write(index + expKey, 0xFF);             // reset key
        EEPROM.write(index + expFeedBack, 0xFF);        // reset key
        // should we reset key too
      }
      switch (i) // reset port
      {
      case 0 ... 7: // Relais Block 1
        pin = i;
        relais_1_Do(pin, 0x00); // turn off
        break;
      case 8 ... 15: // Relais Block 2
        pin = i - 8;
        relais_2_Do(pin, 0x00); // turn off
        break;
      case 16 ... 23: // Relais Block 3
        pin = i - 16;
        relais_3_Do(pin, 0x00); // turn off
        break;
      case 24 ... 31: // Relais Block 4
        pin = i - 24;
        relais_4_Do(pin, 0x00); // turn off
        break;
      }
    }
  }
}
// END portOff

/*
 * uint16_t confLen()
 * calculate length of config, config is terminated wit an 0x55 0xAA value pair
 * config can have a max len of 1024 as we stop counting there
 * but there is a restriction on config to a maximum size of 512 bytes
 * for there are other things stored starting from byte 512
 * if no 0x55 0xAA is found we return zero length
 */
uint16_t confLen()
{
  uint16_t i;
  for (i = 0; i < 512; i++)
  {
    if (EEPROM.read(i) == 0x55)
    {
      if (EEPROM.read(i + 1) == 0xAA)
      {
        return i + 2;
      }
    }
  }
  return 0;
}
// END confLen

/*
 * void do_request()
 * we got a request, see M_Data for details
 * expBase start of array in EEPROM
 */
void do_Request()
{
  uint8_t portState;
  uint8_t port;
  // rxBuf[M_Data] has request
  switch (rxBuf[M_Data])
  {
  // port state requested
  case PortState: // M_Data + 1 will have port number port state request
    portState = EEPROM.read(expBase + (expBaseMult * rxBuf[M_Data + 1]));
    port = rxBuf[M_Data + 1];
    do_Send(2, D_Node, txHead[N_ID], Feedback, port, portState); // send information back
    break;
  // we got a ping
  case PingRequest:
    do_Send(0, D_Node, txHead[N_ID], PingReply); // send a ping reply
    break;
    // more request can be reset etc.
  }
}
// END do_Request

/*
 * void fillLCDline(uint8_t line, uint8_t sign, uint8_t pos)
 * shows state of expader port
 * we show two rows with two 8 char groups
 * Display corresponds to relais blocks
 * line is display row, sign is state, pos is column
 */
void fillLCDline(uint8_t line, uint8_t sign, uint8_t pos)
{
  switch (sign)
  {
  case 0:
    sign = 0x2D; // "-" means OFF
    break;
  case 1:
    sign = 0x2B; // "+" means ON
    break;
  case 2:
    sign = 0x2a; // "*" means time ?
    break;
  default:
    sign = 0x20; // " " undefined
    break;
  }
  lcd.setCursor(pos, line);
  lcd.write(sign);
}
// END fillLCDline

/*
 * void showID(uint8_t ID)
 * uses LEDs to show ID in binary if LEDs exist for a limited time
*/
void showID(uint8_t ID, uint16_t time = 1000)
{
  // switch on LEDs if possible shows node ID in binary
  // LED_1_gn = bit 0, LED_2_gn = bit 1 ..., LED_1_rd = bit 4, LED_2_rd = bit 5 ...
  uint8_t check = 1; // set bit 0
  for (uint8_t i = 0; i < 8; i++)
  {
    showID_LED(ID, i, check, 1);
    check = check * 2; // shift bit left
  }
  delay(time);
  // switch off LEDs
  for (uint8_t i = 0; i < 8; i++)
  {
    showID_LED(ID, i, check, 0);
  }
}
// END showID

/*
 * void showID_LED(uint8_t i, uint8_t check, uint8_t OnOff)
 * switches LEDs on off 
 */
void showID_LED(uint8_t ID, uint8_t i, uint8_t check, uint8_t OnOff)
{
  uint8_t port;
  uint8_t dim;
  // bit 0-3
  if (i < 4) // green led
  {
    port = isLED(i, gnLED);
    dim = isDim(i, gnLED);
  }
  // bit 4-7
  else // red led
  {
    port = isLED(i, rdLED);
    dim = isDim(i, rdLED);
  }
  if (port < 0xFF) // 0xFF means no LED defined
  {
    if (OnOff == 1) // switch LED on if ID bit 1
    {
      if ((ID & check) > 0)
      {
        writeLED(port,0xFF);
      }
    }
    else          // reset all LEDs
    {
      writeLED(port,dim);
    }
  }
}
// end showID_LED

/*
 * uint8_t isLED(uint8_t led, uint8_t color)
 * returns LED port or 0xFF if not configured
 * we support check all max possible 12 LEDs
 */
uint8_t isLED(uint8_t led, uint8_t color)
{
  // offset to ports values 
  led = led % 4;
  return configValue(ledBase, (7 * led) + (2 * color) + 1);
} // end isLED

/*
 * uint8_t isDim(uint8_t led, uint8_t color)
 * returns LED state if Off, my be realy off may be dimmed or on
 * depends on config
*/
uint8_t isDim(uint8_t led, uint8_t color)
{
  led = led % 4;
  return configValue(ledBase, (7 * led) + (2 * color) + 2);
} // end isDim

#ifdef I2C
/*
 * void I2Cscan()
 * option to show I2C addresses
 * my be send via CAN
 */
void I2Cscan()
{
  byte error, address;
  int nDevices;

  Serial.println("Scanning...");

  nDevices = 0;
  for (address = 1; address < 127; address++)
  {
    // The i2c_scanner uses the return value of
    // the Write.endTransmisstion to see if
    // a device did acknowledge to the address.
    Wire.beginTransmission(address);
    error = Wire.endTransmission();

    if (error == 0)
    {
      Serial.print("I2C device found at address 0x");
      if (address < 16)
        Serial.print("0");
      Serial.print(address, HEX);
      Serial.println("  !");

      nDevices++;
    }
    else if (error == 4)
    {
      Serial.print("Unknown error at address 0x");
      if (address < 16)
        Serial.print("0");
      Serial.println(address, HEX);
    }
  }
  if (nDevices == 0)
    Serial.println("No I2C devices found\n");
  else
    Serial.println("done\n");
}
#endif

/* 
 * setPort(uint8_t port, uint8_t portset)
 * finaly set the expand port to new state
 */
void setPort(uint8_t port, uint8_t portset)
{
	uint8_t pin;
  switch (port)
  {
  case 0 ... 7:                // Relais Block 1
    pin = port;                // set pin
    relais_1_Do(pin, portset); // remember state
    break;
  case 8 ... 15: // Relais Block 2
    pin = port - 8;             // calc expander port
    relais_2_Do(pin, portset);  // remember state
    break;
  case 16 ... 23: // Relais Block 3
    pin = port - 16;            // calc expander port
    relais_3_Do(pin, portset);  // remember state
    break;
  case 24 ... 31: // Relais Block 4
    pin = port - 24;            // calc expander port
    relais_4_Do(pin, portset);  // remember state
    break;
  }
}
// END END END END
